#!/usr/bin/perl
#
#  Copyright saturn@surrealchat.net
#  multiple feature-adds and code changes tabris@tabris.net
#
#  Licensed under the GNU Public License 
#  http://www.gnu.org/licenses/gpl.txt
#

package securitybot;

use strict;
no strict "refs";
use FreezeThaw;
use Time::Local;
#use Time::HiRes qw(sleep);

my $debug = 0;
my $master = "saturn";

#this stuff needs to be put into files
my $sbnick = "SecurityBot";
my $ident = 'Security';
my $gecos = 'Security Monitor (you are being monitored)';
my $umodes = '+BHSdopq';
my $vhost = 'services.SC.bot';

my @except = ("localhost", "nascent.surrealchat.net", "cpe-65-30-240-15.mn.rr.com", "192.168.*.*"); #except doesn't work?
my @gexcept = ();

my %ip : shared;
my %flood : shared;
my %conf : shared;
my %drone : shared;
my %ccode : shared;
my %torip : shared;

my %unwhois : shared;

sub init {
	event::addhandler('SEOS', undef, undef, __PACKAGE__, "securitybot::register");
}

sub register {
	net::agent_connect($sbnick, $ident, $vhost, $umodes, $gecos);
	net::sqline($sbnick, 'Reserved for Services');
	event::addhandler('PRIVMSG', undef, $sbnick, __PACKAGE__, "securitybot::msghandle");
	event::addhandler('JOIN', undef, undef, __PACKAGE__, "securitybot::joinhandle");
	event::addhandler('KILL', undef, $sbnick, __PACKAGE__, "securitybot::killhandle");

	loadconf(0);
	loadcountry() if $conf{'EnableCountry'};
	
	if($conf{'EnableTor'} or $conf{'EnableCountry'}) {
		event::addhandler('NICKCONN', undef, undef, __PACKAGE__, 'securitybot::nickconn');
		event::addhandler('USERIP', undef, undef, __PACKAGE__, 'securitybot::userip');
	}
	
	net::join($sbnick, $config::conf{diag});
	net::setmode($sbnick, $config::conf{diag}, '+o', $sbnick);
	my @chans = keys(%drone);
	foreach my $ch (@chans) {
		net::join($sbnick, $ch);
		net::setmode($sbnick, $ch, '+so', $sbnick);
	}
	event::addtimer("securitybot expireflood", 60, __PACKAGE__, "securitybot::expireflood");

	update_tor_list_timed(3540) if $conf{'EnableTor'};
};

sub nickconn {
	my ($rnick, $time, $ident, $host, $vhost, $server, $modes, $gecos) = @_[0,2..4,8,5,7,9];

	if($conf{'EnableTor'} or $conf{'EnableCountry'}) {
		$unwhois{lc $rnick} = 1;
		net::userip($rnick) unless module::is_loaded('services');
	}
}

# XXX - we depend on nickserv to send userip to every user, so this won't work
# without services loaded
sub userip {
	my($src, $rnick, $ip) = @_;

	return unless($unwhois{lc $rnick});
	return unless($ip =~ /^\d{1,3}(\.\d{1,3}){3}$/);

	if($conf{'EnableTor'}) { check_tor($rnick, $ip) or return; }
	check_country($rnick, $ip) if $conf{'EnableCountry'};

	delete $unwhois{lc $rnick};
}

sub check_tor($$) {
	my ($rnick, $ip) = @_;
	
	if($torip{$ip}) {
		net::zline($sbnick, $ip, $conf{'TorZlineTime'}, $conf{'TorZlineReason'});
		return 0;
	}

	return 1;
}

sub check_country($$) {
	my ($rnick, $ip) = @_;
	my @ip = split(/\./, $ip);
	my $rip = join('.', @ip[3,2,1,0]);
	local $/;
	
	open HOST, '-|', "host -W 1 -t txt $rip.zz.countries.nerd.dk";
	my $host = <HOST>; close HOST;

	my $country;
	if($host =~ /text "(\w\w)"/) {
		$country = uc $1;
	} else {
		$country = "Unknown";
	}
	
	my $cname = $ccode{$country};
	$country .= " ($cname)" if $cname;

	net::privmsg($sbnick, $config::conf{diag}, "$rnick is connecting from: $country");
}

sub killhandle($$$$) {
	net::agent_connect($sbnick, $ident, $vhost, $umodes, $gecos);
	net::join($sbnick, $config::conf{diag});
	net::setmode($sbnick, $config::conf{diag}, '+o', $sbnick);
	
	my @chans = keys(%drone);
	foreach my $chan (@chans) {
		net::join($sbnick, $chan);
		net::privmsg($sbnick, $config::conf{diag}, "joining $chan");

	}
}

#sub nmap {
#	my ($ip) = @_;
#	return 0;
#	return if(fork);
#	
#	my $nmap = qx{nmap -P0 $ip};
#	
#	if(
#		$nmap =~ /subseven/
#	) {
#		kline("*\@$ip", "+24h", "Your host is infected with Sub7. Please go to http://www.icstatic.com/content/documents/antivirus/sub7/");
#	}
#
#	exit 0;
#}

sub update_tor_list_timed($) {
	my $time = shift;
	$time = 3600 unless $time;

	event::addtimer('', $time, __PACKAGE__, 'securitybot::update_tor_list_timed');
	
	update_tor_list();
}

sub update_tor_list() {
	net::privmsg($sbnick, $config::conf{diag}, " -- Loading Tor server list.");
	
	open TOR, '-|', ('wget -q -O - ' . $conf{'TorServer'});

	my %newtorip;
	while(my $l = <TOR>) {
		if($l =~ /^router (\S+) (\S+)/) {
			$newtorip{$2} = 1 unless $2 =~ /^(127|10|192\.168)\./;
		}
	}

	close TOR;

	my $torcount = scalar(keys(%newtorip));

	if($torcount > 0) {
		{
			lock %torip;
			%torip = %newtorip;
		}
		net::privmsg($sbnick, $config::conf{diag}, " -- Finished loading Tor server list - $torcount servers found.");
	} else {
		net::privmsg($sbnick, $config::conf{diag}, " -- Failed to load Tor server list, CHECK YOUR TorServer SETTING.");
	}
}

#sub sigalrm {
#	ircsend("PING :$remote");
#	alarm 60;
#}

#sub sigfatal {
#	graceful_die(1); #if $child;
#}

sub graceful_die {
	saveconf();

}

sub expireflood{
	foreach my $isp (keys(%flood)) {
		delete $flood{$isp} if time() - $flood{$isp}[1] > $conf{"FloodGraceTime"};
	}
	event::addtimer("securitybot expireflood", 60, __PACKAGE__, "securitybot::expireflood");
};

sub joinhandle($$) {
	my ($jnick, $chan) = @_;
	$chan = lc($chan);
	my $user = { NICK => $jnick, AGENT => $sbnick };
		
	if ($drone{$chan}) {
		if(!($net::agents{lc $jnick})) {
			net::privmsg($sbnick, $config::conf{diag}, $jnick." joined drone channel ".$chan);
			#unless (adminserv::is_svsop($user, adminserv::S_HELP)) {
				#my $jnickid = nickserv::get_nick_id($user);
				my ($droneident, $dronehost) = nickserv::get_host($user);
				net::zline($sbnick, $dronehost, $conf{"DroneKlineTime"}, $conf{"DroneKlineReason"}.": $chan");
				#net::notice($sbnick, $config::conf{diag}, "Droned: $jnick!$droneident\@$dronehost");
			#}
		}
	}
}

sub msghandle {
		my ($rnick, $dst, $msg) = @_;
		print join("\n", @_);
		my $user = { NICK => $rnick, AGENT => $sbnick };
		unless (adminserv::is_ircop($user)) {
			net::notice($sbnick, $rnick, 'Permission Denied');
			return;
		}

		if($msg =~ /^help/i) {
			net::notice($sbnick, $rnick,
				"- This is SecurityBot.",
				"- ",
				"- Available Commands:",
				"-  JOIN #channel: Join a channel.",
				"-  PART #channel: Leave a channel.",
				"-  NOTICE nick message: Say something.",
				"-  MSG nick message: Say it again.",
				"-  RAW message: Say it from the heart.",
				"-",
				"-  NMAP ip: Scan for open ports and kline",
				"-  TOR-UPDATE: Update Tor server list",
				"-",
				"-  KILL nick reason: Commit murder.",
				"-  KLINE hostmask time reason: Add K:line",
				"-  KLNICK nick: Add K:line for nick",
				"-",
				"-  DRONE #channel: Kline entire channel",
				"-  UNDRONE #channel: Forget kline channel",
				"-",
				"-  FP nick: Manually fingerprint a client",
				"-  FPLIST regexp/number: Display fingerprint information",
				"-  FPRM number: Delete a fingerprint record",
				"-  FPRESET: Flush all fingerprint records",
				"-",
				"-  STATS: View connection statistics.",
				"-  FSTATS: View flooder statistics.",
				"-  RESET: Forget statistics.",
				"-  CONF: View configuration.",
				"-  SET name value: Edit configuration.",
				"-  AUTH hostmask: Allow someone to control me.",
				"-  REHASH: Revert configuration.",
				"-  SAVE: Save configuration.",
				"-  QUIT: Commit suicide.",
				"-  RESTART: Save data and restart.");
			}

			#if($msg =~ /^join (\S*)/i) {
			#	my $chan = lc($1);
			#	net::join($sbnick, $chan);
			#	#net::ircsend(":$sbnick MODE $chan +so $sbnick");
			#	net::setmode($sbnick, $chan, '+so', $sbnick);
			#	push @sitchan, $chan;
			#}

			#if($msg =~ /^part (\S*)/i) {
			#	my $chan = lc($1);
			#	net::part($sbnick, $chan, '');
			#	for(my $i=0; $i<@sitchan; $i++) {
			#		splice(@sitchan, $i, 1) if ($sitchan[$i] eq $chan);
			#	}
			#}

			if($msg =~ /^notice (\S*) (.*)/i) {
				net::notice($sbnick, $1, $2);
			}

			if($msg =~ /^msg (\S*) (.*)/i) {
				net::privmsg($sbnick, $1, $2);
			}

			if($msg =~ /^raw (.*)/i) {
				net::ircsend($1);
			}

			#if($msg =~ /^nmap (.*)/i) {
			#	nmap $1;
			#}

			if($msg =~ /^kill (\S*) (.*)/i) {
				net::irckill($sbnick, $1, $2);
			}

			# Probably should add a parser so it can do ident@host
			# thus allowing it to be like OS AKILL
			#if($msg =~ /^kline (\S*) (\S*) (.*)/i) {
			#	my @t = localtime time;
			#	
			#	net::kline($sbnick, undef, $1, $2, $3 . " (" . $t[3] . "-" . $t[4]  . " $rnick)");
			#}

			#if($msg =~ /^zline (\S*) (\S*) (.*)/i) {
			#	my @t = localtime time;
			#	
			#	net::zline($sbnick, undef, $1, $2, $3 . " (" . $t[3] . "-" . $t[4]  . " $rnick)");
			#}

			#if($msg =~ /^klnick (\S*)/i) {
			#	my ($knick) = ($1);

			#	ircsend("USERHOST :$knick");
			#	push @klq, [$knick, "Requested by an oper"];
			#}

			if($msg =~ /^drone (\S+)/i) {
				my $chan = lc($1);
				
				net::join($sbnick, $chan);
				net::setmode($sbnick, $chan, '+so', $sbnick);
				$drone{$chan} = 1;
			}

			if($msg =~ /^undrone (\S+)/i) {
				my $chan = lc($1);

				net::part($sbnick, $chan, 'undroned by '.$rnick);
				delete $drone{$chan};
			}

			if($msg =~ /^stats/i) {
				my @k = keys(%ip);
				my @v = values(%ip);
				
				net::notice($sbnick, $rnick, "Connection Statistics:");
				for(my $i=0; $i<@k; $i++) {
					net::notice($sbnick, $rnick, $k[$i].", ".$v[$i][0].", ".(time() - $v[$i][1]));
				}
			}

			if($msg =~ /^fstats/i) {
				my @k = keys(%flood);
				my @v = values(%flood);
				
				net::notice($sbnick, $rnick, "Flooder Statistics:");
				for(my $i=0; $i<@k; $i++) {
					net::notice($sbnick, $rnick, $k[$i].", ".$v[$i][0].", ".(time() - $v[$i][1]));
				}
			}

			if($msg =~ /^reset/i) {
				net::notice($sbnick, $rnick, "Resetting.");
				%ip = ();
			}

			if($msg =~ /^conf/i) {
				my @k = keys(%conf);
				my @v = values(%conf);

				net::notice($sbnick, $rnick, "Configuration:");
				for(my $i=0; $i<@k; $i++) {
					net::notice($sbnick, $rnick, $k[$i]." = ".$v[$i]);
				}
			}

			if($msg =~ /^set (\S+) (.*)/i) {
				my @p = ($1, $2);
				chomp $p[1];

				if(update_conf($p[0], $p[1])) {
					net::notice($sbnick, $rnick, "Configuration: ".$p[0]." = ".$p[1]);
				} else {
					net::notice($sbnick, $rnick, "That value is read-only.");
				}
			}

			if($msg =~ /^save/i) {
				net::notice($sbnick, $rnick, "Saving configuration.");

				saveconf();
			}

			if($msg =~ /^rehash/i) {
				net::notice($sbnick, $rnick, "Loading configuration.");

				loadconf(1);
			}

			#if($msg =~ /^quit/) {
			#	ircsend("QUIT :Received QUIT command.");
			#	graceful_die(0);
			#}

			#if($msg =~ /^restart/) {
			#	ircsend("QUIT :Restarting.");
			#	graceful_die(1);
			#}

			#if($msg =~ /^fp (\S+)/) {
			#	fingerprint($1);
			#}

			#if($msg =~ /^fplist (.*)/) {
			#	my $qnick = $1;

			#	for(my $i=0; $i<@fp; $i++) {
			#		next unless $fp[$i]{"lastnick"} =~ /$qnick/i or ($qnick == $i and $qnick =~ /^\d+$/) or $qnick eq "";
			#		
			#		my @k = keys(%{$fp[$i]});
			#		my @v = values(%{$fp[$i]});
			#		
			#		notice($rnick, "- " . $i);
			#		for(my $j=0; $j<@k; $j++) {
			#			if($k[$j] eq "time") {
			#				notice($rnick, $k[$j]." = ".localtime($v[$j])." (".$v[$j].")");
			#			} else {
			#				notice($rnick, $k[$j]." = ".$v[$j]);
			#			}
			#		}
			#	}
			#}

			#if($msg =~ /^fprm (\S+)/) {
			#	splice @fp, $1, 1;
			#}

			#if($msg =~ /^fpreset/) {
			#	undef(@fp);
			#}

			#if($msg =~ /^fpdone (\d+)/) {
			#	kline("*\@".$fp[1]{"userhost"}, 600, "Evasion: ".$fp[1]{"lastnick"}) if(fpmatch($fp[$1]));
			#	ircsend("PRIVMSG #Services :\001ACTION detected a possible evader: ".$fp[$1]{"lastnick"}."\001") if(fpmatch($fp[$1]));
			#}

			if($msg =~ /^tssync/i) {
				net::tssync();
			}
			if($msg =~ /^svsnick (\S+) (\S+)/i) {
				net::svsnick($sbnick, $1, $2);
			}
			if($msg =~ /^tor-update/i) {
				net::notice($sbnick, $rnick, "Updating Tor server list.");
				update_tor_list();
			}

		#if($in =~ /:.*(PRIVMSG|NOTICE) \Q$sbnick\E\s*:(.*)/i) {
		#	my $msg = $2;

			#if($msg =~ /KILL message for (.+?)\!.+?\@.*admin\.\S+?\.ops from (.+?) /) {
			#	irckill($sbnick, $2, $1) unless($2 eq $sbnick or $2 eq "services.webchatting.com");
			#	notice($conf{"Debugger"}, "KILLKILL $2 $1") if $conf{"Debug"};
			#}
			
			#if($msg =~ /^\*\*\* Notice -- Client (.*?): (\S+?) \((.+?)\@(.+?)\)/) {
			#	my ($act, $rnick, $id, $ip) = ($1,$2,$3,$4);
			
		#	if($msg =~ /^\*\*\* Notice -- Client (.*?): /) {
		#		my $act = $1;
		#		my ($rnick, $id, $ip);

		#		if($msg =~ /(\S+?)\!(\S+?)\@(\S+?) /) {
		#			($rnick, $id, $ip) = ($1,$2,$3);
		#		} else {
		#			$msg =~ /: (\S+?) \((.+?)\@(.+?)\)/;
		#			($rnick, $id, $ip) = ($1,$2,$3);
		#		}

		#		$id=lc($id); $ip=lc($ip);

		#		notice($conf{"Debugger"}, $act, $rnick, $id, $ip) if $conf{"Debug"};
				
		#		if($conf{"EnableConnTrack"} and $act =~ /connecting/) {
					#foreach my $ex (@except) { next LINE if $ex eq $ip; }
		#			my $ot = $ip{$ip}[1];
		#			$ip{$ip}[0]++;
		#			$ip{$ip}[1] = time;
		#			#$ip{$ip} = 0 if($ip{$ip} < 0);

		#			notice($conf{"Debugger"}, "$ip ".$ip{$ip}[0], "$act $rnick $id") if $conf{"Debug"};
				
		#			if((($ip{$ip}[1] - $ot) <= $conf{"MaxCloneTime"}) and ($ot != 0)) {
		#				foreach my $ex (@gexcept) { if($ip =~ /\Q$ex\E$/) {
		#					irckill(,$sbnick, $rnick, $conf{"CloneKillReason"});
		#					next LINE;
		#				} }
		#				kline("*\@$ip", $conf{"CloneKlineTime"}, $conf{"CloneKlineReason"});
		#			}

		#			if($ip{$ip}[0] > $conf{"MaxClones"}) {
		#				ircsend("KILL $rnick :".$conf{"CloneKillReason"});
		#			}

		#			if ($conf{"EnableAutoFP"}) {
		#				fingerprint($rnick);
		#			}
		#		}

		#		if($conf{EnableConnTrack} and $act =~ /exiting/) {
		#			#foreach my $ex (@except) { next LINE if $ex eq $ip; }
		#			$ip{$ip}[0]--;
		#			delete $ip{$ip} if $ip{$ip}[0] <= 0;
		#		}
		#	}

		#	if($conf{"EnableFloodTrack"} and ($msg =~ /Flood -- \S+\@(\S+)/
		#	or $msg =~ /\(\S+\@(\S+)\) \[Killed \(OperServ/)) {
		#		my $ip = $1;

		#		notice($conf{"Debugger"}, "$ip Flood") if $conf{"Debug"};

		#		my $isp;
		#		if($ip =~ /\d+\.\d+\.\d+\.\d+/) {
		#			$ip =~ /.+? \. (.+)$/x;
		#			$isp = $1;
		#		} else {
		#			$ip =~ /\. (\w+? \. \w+?)$/x;
		#			$isp = $1;
		#		}

		#		delete $flood{$isp} if time() - $flood{$isp}[1] > $conf{"FloodGraceTime"};

		#		$flood{$isp}[0]++;
		#		if($flood{$isp}[0] >= $conf{"FloodGrace"} and time() - $flood{$isp}[1] < $conf{"FloodGraceTime"}) {
		#			kline("*\@$ip", $conf{"FloodKlineTime"}, $conf{"FloodKlineReason"});
		#		}
				
		#		$flood{$isp}[1] = time;
		#	}
		#
		#}
}

sub saveconf() {
	config::writeHash(\%conf, "config/securitybot/sb.conf");
	config::dsave(\%ip, "config/securitybot/sb.ip");
	config::dsave(\%flood, "config/securitybot/sb.flood");
	config::writeHash(\%drone, "config/securitybot/sb.drone");
}

sub loadconf($) {
	my ($update) = @_;
	
	%conf = config::readHash("config/securitybot/sb.conf");
	%ip = %{ config::dload("config/securitybot/sb.ip") };
	%flood = %{ config::dload("config/securitybot/sb.flood") };
	%drone = config::readHash("config/securitybot/sb.drone");
}

sub update_conf($$) {
	my ($k, $v) = @_;

	return 0 if($k eq 'EnableTor' or $k eq 'EnableCountry');

	$conf{$k} = $v;
	return 1;
}

sub loadcountry() {
	open COUNTRY, "config/securitybot/country-codes.txt";
	while(my $x = <COUNTRY>) {
		chomp $x;
		my($code, $country) = split(/   /, $x);
		$ccode{uc $code} = $country;
	}
	close COUNTRY;
}

sub begin { }
sub end { }
sub unload { saveconf(); }

1;
